
/* ZELLER.C  Finds day of week or date of Easter */
/* Uses the algorithm given by Ch. Zeller, Acta Mathematica, v 9, 1897 */

/*
    Uses Julian or Gregorian calendar, as desired.
    Range 1 AD to 4000 AD.

    Dates may be entered numerically (e.g. 1/2/1964 or 1-2-1964 or 1 1 1964...
    or in English: Jan 2, 1964 or 4 July 1776 for example (you must enter the
    first three digits of the month name correctly, or it will be ignored;
    moverover, the day must come before the year.)

    The year must always be given, and must be entered in full. The
    month or day may be omitted, however, with special significance:
      If only the year is entered, the date of Easter will be displayed.
      If only the month and year are entered, a calendar for that month
      will be displayed. (PgUp and PgDn for previous/next month.)

    Normally, the Gregorian calendar will be used up to Oct 15, 1582, but this
    can be changed by adding a letter at the end of the year:
       G to force Gregorian mode
       J to force Julian mode
       B to force British style (i.e.Julian calendar ends 9/2/1752, Gregorian
         starts 9/14/1752)

   Astronomers start the Gregorian calendar on Oct 15, 1582, the earliest
   date of its adoption anywhere. Historically, however, most of the world
   waited quite some time after this before making the change - Britain
   in 1752, Russia in 1918, Greece in 1923, .... Consequently, to determine
   the day of the week for a given historical date, one must know the country
   of origin. For those countries which adopted the Gregorian calendar on
   Oct 15, 1582 (Italy, Spain, Portugal and Poland), or for any date which
   precedes this earliest changeover point, one can use the Julian Day (JD)
   of astronomy to find the day of the week: (JD + 1) % 7 = 0 for Sunday,
   1 for Monday,... 6 for Saturday. For other dates, this formula should not
   be used since the astronomical Julian day uses Oct 15, 1582 as the first
   day of the Gregorian calendar. For positive years, the Zeller code given
   here will find the day of the week for any month/day in either calendar.
   In his paper cited earlier, Zeller also gives an algorithm for determning
   the date of Easter, but it is in error for some years (e.g. 1954, 1981).
   I have consequently used the code due to Gauss and explained in Knuth,
   The Art of Computer Programming, vol 2, pp. 212-220.
*/

#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <dos.h>
#include <stdarg.h>

int Normal,Gregorian,British; /* flags: Normal -  Julian to Oct 4, 1582 */
                              /*        British - Julian to Sep 2, 1752 */
int mono, norm;
int leap;
/*
   In the Julian calendar, the year Y is a leap year if it is divisible by 4.
   In the Gregorian calendar, all multiples of 4 are leap years except those
   which are multiples of 100 but not multiples of 400. It has been proposed
   that the Gregorian rule be changed so that years which are multiples of
   4000 will not be leap years.
*/

/* day_tab is copied from K&R, for use in month_day() and day_of_year(), which
   also came from K&R; it is not accurate for 1582, or the year of change to
   the Gregorian calendar if that is not 1582 (e.g. in the normal usage,
   October, 1582 had only 21 days, and in British usage, September, 1752 had
   only 19 days; similar differences occur in other countries). */
int day_tab[2][13] = {  /* day_tab[leap][month] = number of days in 'month' */
     {0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
     {0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31}};

main()
{
  int month,day,year,wk;  /* wk = day of week: 0=Sat,...,6=Fri */
  char *day_name[] = {
    "Saturday","Sunday","Monday","Tuesday","Wednesday","Thursday","Friday"};
  char s[32];             /* for user input */
  int vi[4];              /* for user input */
  char Easter[10];        /* date string */
  struct text_info oldtext;  /* for storing original text settings */
  int oldattr;            /* original attribute- restore on exit */
  int i, y, row, col;

  clrscr();
  gettextinfo(&oldtext);
  oldattr = oldtext.normattr;
  if (oldtext.currmode == '\7') {   /* assume mono card if mode 7 on startup */
    mono = 1;
    norm = 7;
  }
  entry_screen();          /* basic screen */
  window(10,8,66,23);      /* entry window */
  if (mono)
    textattr(norm);        /* use DOS drab on mono systems */
  else
    textattr(31);          /* blue background, bright white foreground */
  clrscr();
  row = 1; col = 13;       /* initial entry position */
  while(1) {
    s[0] = '\0';            /* set original entry string to "" */
    if (row >= 17) {        /* start over when window is full */
      clrscr();
      row = 1;
    }
    gotoxy(col,row++);
    month = day = year = 0;
    cprintf("   Date: ");
    if (getstr(s) == -1)  {  /* F1 pressed */
      show_adoptions();      /* display table of Gregorian adoptions */
      row--;                 /* reset row */
      continue;
    }
    if (!s[0])                /* terminate if empty string entered */
      break;
    if (!read_date(s,&month,&day,&year)) { /* read_date() tries to decipher s */
      error_msg(col,row++,"Illegal date");
      continue;
    }
    if (year < -100 || year > 3000) {
      error_msg(col,row++,"Illegal date");
      continue;
    }
    if (Normal) {   /* no Gregorian dates before Oct 15, 1582, anywhere */
      if ( (year<1582) || (year==1582 && month<10) ||
                        (year==1582 && month==10 && day<15))
        Gregorian = 0;
      else
        Gregorian = 1;
      }
    else
     if (British) {  /* no Gregorian dates before Sep 14, 1752 in Britain */
       if ( (year<1752) || (year==1752 && month<9) ||
                        (year==1752 && month==9 && day<14))
         Gregorian = 0;
       else
         Gregorian = 1;
     }
    y = year;
    if (Gregorian)
      leap = (y%4 == 0 && y%100 != 0 || y%400 == 0);
    else
      leap = (y%4 == 0);

    if (month == 0 && day == 0) {  /* give date of Easter for 'year' */
      easter(year,&month,&day,Easter);
      gotoxy(col,row++);
      cprintf(" Easter: %s",Easter);
    }
    else {
      if (month<1 || month>12 || day<0 || day>day_tab[leap][month]) {
        error_msg(col,row++,"Illegal date");  /* day_tab = num days in month */
        continue;
      }
      if (day == 0)     /* no day (or zero day) entered - draw calendar */
        show_cal(month,year);
      else {
        wk = weekday(month,day,year); /* normal operation - show day of week */
        gotoxy(col,row++);
	cprintf("Weekday: %s",day_name[wk]);
      }
    }
  }
  window(1,1,80,25);           /* reset screen and exit */
  textattr(oldattr);
  clrscr();
}


/* ------------------------------------------------------------------------ */
entry_screen()
{
  int i;
  int left1=9,top1=2,right1=67,bottom1=7;
  char *msg2[] = {
          "     Enter month, day, and year for day of week.      ",
          "     Enter month and year alone for calendar.         ",
          "     Enter year alone for date of Easter.             ",
          "     Enter a blank date to terminate.                 "};

char *footer =
"Press F1 for a list of dates of some adoptions of the Gregorian calendar.";

   window (1,1,80,25);
   textattr(7);
   clrscr();
   if (!mono)
     textattr(14);
   gotoxy(34,1);
   cputs("ZELLER");
   if (mono) {
     rect(left1,top1,right1,24,norm,1);
     textattr(norm);
   }
   else {
     rect(left1,top1,right1,24,16*1+15,1);
     textattr(31);
   }
   /* draw crossbar below instructions */
   gotoxy(left1,bottom1);
   putch(195);
   for (i=left1+1; i<right1; i++)
    putch(196);
   putch(180);
   window(left1+1,top1+1,right1-1,bottom1-1);
   clrscr();
   for (i=0; i<4; i++) {
     gotoxy(2,1+i);
     cputs(msg2[i]);
   }
   window(1,1,80,25);
   textattr(7);
   gotoxy(2,25);
   cputs(footer);
}

/*---------------------------------------------------------------------------*/
/* decipher user input for date */
read_date(char *s, int *m, int *d, int *y)
{
  static char *month_name[] =
  {"JAN","FEB","MAR","APR","MAY","JUN","JUL","AUG","SEP","OCT","NOV","DEC"};
  int cnt,vi[3] = {0,0,0};
  int i,month=0;

  strupr(s);
  i = strlen(s)-1;
  while (s[i] == ' ') i--;  /* remove trailing spaces, if any */
  s[i+1] = '\0';
  if (s[i] == 'J') { /* if last character is J, use Julian calendar */
    Gregorian = 0;
    British = 0;
    Normal = 0;
  }
  else
  if (s[i] == 'G') { /* if last character is G, use Gregorian calendar */
    Gregorian = 1;
    British = 0;
    Normal = 0;
  }
  else
  if (s[i] == 'B') { /* if last character is B, use British calendar */
    Gregorian = 0;
    British = 1;
    Normal = 0;
  }
  else {      /* use Julian to Oct 4, 1582, Gregorian after Oct 14, 1528 */
    Gregorian = 0;
    British = 0;
    Normal = 1;
  }
  /* for existence of month name in s */
  for (i=0; i<12; i++)
    if (strstr(s,month_name[i]))
      break;
  if(i < 12)
    month = i+1;      /* if name is given, set month appropriately */
  cnt = getint(s,&vi[0],&vi[1],&vi[2]);
  if (cnt == 0)          /* no integers in s */
    return 0;
  if (cnt == 1) {        /* only one integer; must be the year */
    *m = month;          /* month may have been given by name */
    *d = 0;
    *y = vi[0];
    return 1;
  }
  if (cnt == 2) {  /* only two integers; month must have been entered by name */
    *m = month;
    *d = vi[0];
    *y = vi[1];
    return 1;
  }
  if (cnt == 3) {          /* three integers; must be month,day,year */
    *m = vi[0], *d = vi[1], *y = vi[2];
    return 1;
  }
}
/*---------------------------------------------------------------------------*/
/* weekday() is from Zeller's paper, cited earlier */
weekday(int month, int day, int year)
{
  int J,K,m=month,q=day,h;

  if (m < 3) {
    m += 12;
    year--;
  }
  J = year/100;
  K = year%100;
  h = (q + ((m+1)*26)/10 + K + K/4);      /* Zeller's code */
  if (Gregorian)
    h += (J/4 - 2*J);
  else
    h += (5 - J);
  h = mod(h,7);
  return h;
}

/*---------------------------------------------------------------------------*/
/* The computer science version of 'integer division' is different from that
   used by Zeller (and all other mathematicians) and therefore so is the mod
   function. This is a correct version of mod, needed whenever x might be
   negative. (it is well known that this code could be replaced by
   'x%y+(x%y<0)*y'; it seems preferable to write out 'x mod y' the way '%'
   should have been implemented; the code works just as well if long ints
   are used instead of ints) */
mod(int x, int y)
{
  if (y <= 0)
    return -1;             /* error: y must be positive */
  while (x >= y) x -= y;
  while (x < 0) x += y;
  return x;
}

/*---------------------------------------------------------------------------*/
/* easter() is from Knuth, cited earlier and credited to Gauss */
easter(int year,int *month, int *day, char *Easter)
{
  int a,b,c,C,M,L,d,e,f;
  int Mar21,e_date;

  a = year%4;
  b = year%7;
  c = year%19;
  C = year / 100;
  if (Gregorian) {
    M = 15 + C - C/4 - (8*C+13)/25;
    L = 4 + C - C/4;
  }
  else {
    M = 15;
    L=6;
  }
  d = mod(19*c+M,30);
  f = mod(2*a + 4*b + 6*d + L,7);
  if (d == 29 && f == 6)
    e = d+f-6;
  else
    if (d == 28 && f == 6) {
      if(M==2 || M==5 || M==10 || M==13 || M==16 || M==21 || M==24 || M==29)
        e = d+f-6;
    }
  else
    e = d+f+1;
  Mar21 = day_of_year(3, 21);   /* year already set */
  e_date = Mar21+e;
  month_day(e_date, month, day, Easter);
}

/*---------------------------------------------------------------------------*/
/* from K&R */
day_of_year(int month, int day)  /* assumes 'leap' already set */
{
  int i;

  for (i=1; i<month; i++)
    day += day_tab[leap][i];
  return day;
}

/*---------------------------------------------------------------------------*/
/* essentially from K&R */
/* set month and day, fill s with <monthname[month] year> */
/* assumes the global variable 'leap' has already been set for the year used */
month_day(int day_num, int *month, int *day, char *s)
{
  static char *month_name[] =
   {"","Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"};
  int i;

  for (i=1; day_num > day_tab[leap][i]; i++)
    day_num -= day_tab[leap][i];
  sprintf(s,"%s %-2d",month_name[i],day_num);
  *month = i;
  *day = day_num;
}

/* -------------------- getint(s,...) -------------------------------- */
/*
   Extract integers from string s, put them into specified variables.
   Arguments other than the string should be pointers to ints, e.g.
   getint(s,&x1,&x2,&x3), with x1,x2,x3 having type int. Assumes numbers
   in s are written in standard ASCII form: no isolated or repeated minus
   signs, no decimals, no spaces inside numbers, no exponents, etc. The
   string s need not consist of numbers alone, and pairs of numbers may be
   separated by any non-numerical characters.
*/

int
getint(char *t, ...)
{
  int cnt=0;
  char *p,s[80];
  int *arg;
  va_list ap;

  s[0] = ' '; s[1] = '\0';
  strcat(s,t);   /* be sure string starts with a space so *(p-1) can be used */
  va_start(ap,s);
  p = s;
  while ((arg = va_arg(ap, int *)) != NULL)
    {
     /* move past invalid numerical characters */
a:
     while (*p && !isdigit(*p) && *p != '-') p++;
     if (!*p)
       break;
     /* minus signs must be preceded by a space, followed by a digit */
     if (*p == '-') {
       if (*(p-1) != ' ' || !isdigit(*(p+1))) {
         p++;
         goto a;
       }
     }
     *arg = atoi(p);
     if (*p == '-') p++;                 /* skip leading minus sign */
     while(isdigit(*p)) p++;             /* move through remaining valids */
     cnt++;                              /* finished with current number */
    }
  return cnt;
}

/*------------------------ getstr(char *s) ----------------------------------*/
/* Input characters into string s[].
   Replacement for gets(), which does not echo input in
   current screen background and text color. (Neither does
   cgets().) Input starts at the current cursor position.
   Input characters may be entered from there through col 80,
   same row. Backspace removes the preceding character, but
   no other control keys are recognized.
   Terminate with CR to retain string or with Esc to ignore the
   input characters and set s = "". Returns -1 if F1 pressed.
*/

#define BK  8     /* backspace */
#define CR  13
#define ESC 27
#define F1 -59

int getstr(char *s)
{
  int i,col,row,col0;
  int k,done=0;
  char ch;

  col = col0 = wherex();
  row = wherey();
  if (col0 == 80)
    return 0;
  while(1) {
    gotoxy(col,row);
    switch(k=getkey()) {
      case BK:             /* backspace */
        if (col > col0) {
          col--;
          gotoxy(col,row);
          cputs(" ");
        }
        break;
      case CR:
        done = 1;
        break;
      case ESC:
        s[0] = '\0';       /* set s = "" **/
        return 0;
      case F1:
        return -1;
      default:
        if (isprint(k) && col < 80) {
          putch(k);        /* overwrite current character, if any */
          col++;
          }
    }
    if (done) {
      read_str(s,col0,col,row);   /* read characters on screen into s */
      return 1;
      }
  }
}

/*----------- part of getstr() --------------------*/
/* read character on screen at (col,row) */
read_char(int col, int row)
{
  gotoxy(col,row);
  _AH = 8;
  _BH = 0;
  geninterrupt(0x10);
  return _AL;
}

/*----------- part of getstr() --------------------*/
/* read chars from (col0,row)-(col,row) into s */
read_str(char *s, int col0, int col, int row)
{
  int c;
  char *p = s;

  for (c=col0; c<col; c++)
    *p++ = read_char(c,row);
  *p = '\0';
  gotoxy(col,row);
}

/*---------------------------------------------------------------------------*/
error_msg(int col,int row,char *msg)
{
  gotoxy(col,row);
  cputs(msg);
}

/*---------------------------------------------------------------------------*/
getkey()       /* basic inkey function */
{
  int k;

  k = getch();
  if (k == 0)
    k = -getch();  /* make extended codes negative to distinguish from normal */
  return k;
}

/*---------------------------------------------------------------------------*/
/* rect() - draw rectangle on text screen */
int tlc[3]={0,218,201};
int trc[3]={0,191,187};
int blc[3]={0,192,200};
int brc[3]={0,217,188};  /* single,double corners */
int hl[3]={0,196,205};
int vl[3]={0,179,186};   /* single,double horizontal, vertical lines */

rect(int left, int top, int right, int bottom, int attr, int border)
{
  int i;
  int width, height;

  width = right-left;
  height = bottom-top;

  textattr(attr);
  gotoxy(left,top);
  putch(tlc[border]);
  for (i=1; i<width; i++)
    putch(hl[border]);
  putch(trc[border]);
  for (i=0; i<height-1; i++) {
    gotoxy(left,top+1+i);
    putch(vl[border]);
    gotoxy(left+width,top+1+i);
    putch(vl[border]);
  }
  gotoxy(left,top+height);
  putch(blc[border]);
  for (i=1; i<width; i++)
    putch(hl[border]);
  putch(brc[border]);
}

/*---------------------------------------------------------------------------*/
show_adoptions()
{
  int i;
  char *adoptions[] = {
          " Some dates of adoption of the  Gregorian calendar. ",
          "",
          "Italy, Spain, Portugal    Oct 1582  Omitted: Oct 5-14 ",
          "France                    Dec 1582                    ",
          "Catholic German states        1583                    ",
          "Protestant German states      1700                    ",
          "Great Britain             Sep 1752  Omitted: Sep 3-13 ",
          "Sweden                        1753                    ",
          "Japan                         1873                    ",
          "China                         1912                    ",
          "Russia                        1918                    ",
          "Greece                        1924                    ",
          "Turkey                        1927                    "};

  int left1=9,top1=5,right1=67,bottom1=19;
  int attr;
  char buffer[4000];

  window(1,1,80,25);
  gettext(1,1,80,25,buffer);
  textattr(7);
  clrscr();
  attr = mono ? norm : 31;
  rect(left1,top1,right1,bottom1,attr,1);
  window(left1+1,top1+1,right1-1,bottom1-1);
  textattr(attr);
  clrscr();
  for (i=0; i<13; i++) {
    gotoxy(3,1+i);
    cputs(adoptions[i]);
  }
  window (1,1,80, 25);
  textattr(7);
  gotoxy(2,25);
  cputs("Press any key to return...");
  getch();
  puttext(1,1,80,25,buffer);
  window(10,8,66,23);
  if (mono)
    textattr(norm);
  else
    textattr(31);
}

/*---------------------------------------------------------------------------*/
/* show-cal() - draw calendar for specified month,year */

typedef unsigned char uchar;

/* single box characters */
   /* corners */
#define TLC 218
#define TRC 191
#define BLC 192
#define BRC 217
  /* horizontal, vertical segments */
#define HL  196
#define VL  179
  /* junctions: top-down, bottom-up, left-right, right-left, cross */
#define TDJ 194
#define BUJ 193
#define LRJ 195
#define RLJ 180
#define XX  197
  /* space */
#define sp  32

uchar top_row[] =
  {TLC,HL,HL,HL,HL,TDJ,HL,HL,HL,HL,TDJ,HL,HL,HL,HL,TDJ,HL,HL,HL,HL,
           TDJ,HL,HL,HL,HL,TDJ,HL,HL,HL,HL,TDJ,HL,HL,HL,HL,TRC,0};
uchar ins_row[] =
  {VL,sp,sp,sp,sp,VL,sp,sp,sp,sp,VL,sp,sp,sp,sp,VL,sp,sp,sp,sp,VL,
           sp,sp,sp,sp,VL,sp,sp,sp,sp,VL,sp,sp,sp,sp,VL,0};
uchar crs_row[] =
  {LRJ,HL,HL,HL,HL,XX,HL,HL,HL,HL,XX,HL,HL,HL,HL,XX,HL,HL,HL,HL,XX,
           HL,HL,HL,HL,XX,HL,HL,HL,HL,XX,HL,HL,HL,HL,RLJ,0};
uchar bot_row[] =
  {BLC,HL,HL,HL,HL,BUJ,HL,HL,HL,HL,BUJ,HL,HL,HL,HL,BUJ,HL,HL,HL,HL,
           BUJ,HL,HL,HL,HL,BUJ,HL,HL,HL,HL,BUJ,HL,HL,HL,HL,BRC,0};

char *legend = "  S    M    T    W    T    F    S";

show_cal(int month, int year)
{
  int x0,y0,x,y;
  int i,j,k;
  char s[32];
  int c[42], r[42];
  int first_day, month_length;
  int num_rows;
  static char *month_name[] =
  {"","January","February","March","April","May","June","July","August",
   "September","October","November","December"};
  char buffer[4000];            /* for saving current screen */

  window(1,1,80,25);
  gettext(1,1,80,25,buffer);
a:
  if (mono)
    textattr(norm);
  else
    textattr(31);
  clrscr();
  first_day = (weekday(month,1,year)+6)%7;  /* 0 = Sun, ... 6 = Sat */
  month_length = day_tab[leap][month];
  x = x0 = 20, y = y0 = 7;
  for(i=0; i<42; i++) {
    c[i] = x0 + 5*(i%7) + 2;
    r[i] = y0 + 2*(i/7) + 1;
  }
  if (first_day == 0 && month_length <= 28)
    num_rows = 4;
  else if (first_day + month_length > 35)
    num_rows = 6;
  else
    num_rows = 5;
  gotoxy(x0+14-(strlen(month_name[month])+5)/2,y-3);
  cprintf("%s %d ",month_name[month],year);
  if(Normal && year == 1582)
    if(month <= 10) {
    cputs("(Julian)");
    goto b;
  }
  else {
    cputs("(Gregorian)");
    goto b;
  }
  if(British && year == 1752)
    if(month < 9) {
      cputs("(Julian)");
      goto b;
   }
   else {
     cputs("(Gregorian)");
     goto b;
   }
  if (Gregorian)
    cputs("(Gregorian)");
  else
    cputs("(Julian)");
b:
  gotoxy(x,y-1);
  cputs(legend);
  gotoxy(x,y++);
  cputs(top_row);
  for (i=1; i<num_rows; i++) {
    gotoxy(x,y++);
    cputs(ins_row);
    gotoxy(x,y++);
    cputs(crs_row);
  }
  gotoxy(x,y++);
  cputs(ins_row);
  gotoxy(x,y);
  cputs(bot_row);
  if (!mono)
    textattr(30);
  for (i=first_day,j=1; i<first_day+month_length; i++,j++) {
    gotoxy(c[i],r[i]);
    cprintf("%2d",j);
  }
  if (!mono)
    textattr(23);
  gotoxy(2,24); clreol();
  if (month > 1 && month < 12)
     cputs("PgUp/PgDn for previous/next month, Enter to return");
   else if (month == 1)
     cputs("PgDn for next month, Enter to return");
   else
     cputs("PgUp for previous month, Enter to return");
  while (1) {
    switch(k = getkey()) {
      case -73:  /* PgUp */
        if (month > 1) {
          month--;
          goto a;
        }
        break;
      case -81:  /* PgDn */
        if (month < 12) {
          month++;
          goto a;
        }
        break;
      case CR:   /* Enter or Esc returns */
      case ESC:
        break;
    }
    if (k == CR || k == ESC)
      break;
  }
  puttext(1,1,80,25,buffer);
  window(10,8,66,23);
  if (mono)
    textattr(norm);
  else
    textattr(31);
}

